! ===========================================================================
MODULE act_likelihood_2008

! Parameters are defined in ACT_options module
! Based on WMAP likelihood function
! ===========================================================================

  USE ACT_options
  use act_beam_chisq
  
  character(len=*), parameter, public :: act_likelihood_version='v2a1s'
  
  REAL(8) ::  btt_dat(0:bmax0-1),btt_var(0:bmax0-1)
  REAL(8) ::  fisher(0:bmax0-1,0:bmax0-1),bval(0:bmax0-1)
  REAL(8), dimension (:), allocatable :: cl_sz, cl_c, cl_p, blmin, blmax
  REAL(8), dimension(:,:), allocatable :: win_func
  
  PRIVATE
  public :: act_likelihood_init
  public :: act_likelihood_compute
  public :: get_free_lun
contains
  
  ! ===========================================================================
  SUBROUTINE act_likelihood_init
    ! ===========================================================================
    
    IMPLICIT NONE
    
    INTEGER  :: i,j,lun,il
    REAL(8)  :: laux,llaux, dummy
    CHARACTER(LEN=240) :: ttfilename, bblfilename, binfilename
    LOGICAL  :: good
    
    allocate(blmax(0:bmax0-1), blmin(0:bmax0-1))
    allocate(cl_sz(2:tt_lmax), cl_p(2:tt_lmax),cl_c(2:tt_lmax))

    print *, 'Initializing ACT likelihood, version '//act_likelihood_version
    
    !-----------------------------------------------
    ! set file names
    !-----------------------------------------------
    
    ttfilename  = trim(ACT_data_dir)//'Data/act_spectrum_148x148_v2a1s.dat'
    binfilename = trim(ACT_data_dir)//'Data/act_bins_v2a1s.dat'
    bblfilename = trim(ACT_data_dir)//'Data/act_BblMean_148x148_v2a1s.dat'

    !-----------------------------------------------
    ! load spectrum and band power window functions 
    !-----------------------------------------------
    
    inquire(file=ttfilename,exist = good)
    if(.not.good)then
       write(*,*) 'cant find', trim(ttfilename), trim(ACT_data_dir)
       stop
    endif
    call get_free_lun( lun )
    open(unit=lun,file=ttfilename,form='formatted',status='unknown',action='read')
    

    fisher=0.d0
    do i=0,bmax0-1
       read(lun,*) bval(i),btt_dat(i), btt_var(i)
       btt_var(i)= (btt_var(i))**2 
       fisher(i,i)= 1.d0/btt_var(i)
    enddo
    close(lun)
    
    !Get bin range to be used, using lrange set in Options file
    do i=0,bmax0-1
       if(bval(i) <= lmin) bmin=i+1
       if(bval(i) <= lmax) bmax=i
    enddo
    
    call get_free_lun( lun )
    open(unit=lun,file=binfilename,form='formatted',status='unknown',action='read')
    
    do il = 0,bmax0-1
       read(lun,*) blmin(il), blmax(il), dummy
    end do
    close(lun)
    
    inquire (file=bblfilename,exist = good)
    if(.not.good)then
       write(*,*) 'cant find', trim(bblfilename), trim(ACT_data_dir)
       stop
    endif
    call get_free_lun( lun )
    open(unit=lun,file=bblfilename,form='formatted',status='unknown',action='read')
    
    allocate(win_func(0:bmax0-1,1:tt_lmax))
    do il = 1, tt_lmax
	read(lun,*) i, (win_func(i,il), i=0,bmax0-1) 
	enddo	
    close(lun)

   
    !-----------------------------------------------
    ! initialise beam uncertainties
    !-----------------------------------------------

    if (use_act_beam) call init_act_beam_chisq(win_func,bval,blmin,blmax)

    call get_free_lun( lun )
    open(unit=lun,file=trim(ACT_data_dir)//trim(szfilename),form='formatted',status='unknown')
    do il=2,tt_lmax
       read(lun,*) dummy,cl_sz(il)
    enddo
    close(lun)

    call get_free_lun( lun )
    open(unit=lun,file=trim(ACT_data_dir)//trim(srcfilename),form='formatted',status='unknown')
    do il=2,tt_lmax-1
       read(lun,*) dummy,cl_c(il)
    enddo
    close(lun)
    
    do il=2,tt_lmax
       cl_p(il)=(il/3000.d0)**2 !Normalize Poisson power at l=3000 
    enddo
    
  END SUBROUTINE act_likelihood_init
  
  ! ===========================================================================
  
  SUBROUTINE act_likelihood_compute(cltt,amp_sz,amp_c,amp_p,like)
    
    ! ===========================================================================
    
    IMPLICIT NONE
    REAL(8), intent(in) :: cltt(2:*), amp_sz, amp_c,amp_p
    REAL(8), intent(out) :: like(num_ACT)
    INTEGER :: lun,il, ill ,i   
    REAL(8) :: cltt_temp(2:tt_lmax)
    REAL(8) :: btt_th(0:bmax0-1),diffs(0:bmax0-1),tmp(0:bmax0-1)
    REAL(8) :: dlnlike,denom, chi2op,chi2pp,wpp,wdd,dlnlike_beam 
    
    Like = 0.d0

    !Sum model in B_ell = l(l+1)Cl/2pi, CMB+SZ+ptsrc (Poisson + clustered)
    do il=2,tt_lmax
       cltt_temp(il)=cltt(il) +amp_c*cl_c(il) +amp_p*cl_p(il) +amp_sz*cl_sz(il)
    enddo
    
    call get_free_lun(lun)
    
    !convert back to C_ell
    do il=2,tt_lmax
       cltt_temp(il)=cltt_temp(il)*2.d0*PI/(dble(il)*(dble(il)+1.d0)) 
    enddo

    !Multiply by bandpower matrix
    btt_th(0:bmax0-1)=MATMUL(win_func(0:bmax0-1,2:tt_lmax),cltt_temp(2:tt_lmax))
 
    !Basic chisq 
    diffs(bmin:bmax) = btt_dat(bmin:bmax) - btt_th(bmin:bmax)

    dlnlike = SUM(diffs(bmin:bmax)**2/btt_var(bmin:bmax)) 

    !Calibration error, following cosmomc implementation
    denom=1.d0
    if (calib_err > 1e-4) then
       tmp(bmin:bmax) = btt_th(bmin:bmax)/btt_var(bmin:bmax)
       chi2op = SUM(diffs(bmin:bmax)*tmp(bmin:bmax))
       chi2pp = SUM(btt_th(bmin:bmax)*tmp(bmin:bmax))
       wpp = 1/(chi2pp+1/calib_err**2) 
       dlnlike = dlnlike - wpp*chi2op**2
       denom = denom/wpp*calib_err**2
       dlnlike = dlnlike + log(denom)
    end if
 
    Like(ttlike) = dlnlike/2.d0
    
    !---------------------------------------------------------------------------
    !Beam error 
    !---------------------------------------------------------------------------

    if(use_act_beam)then
       dlnlike_beam = compute_act_beam_chisq(bmin,bmax,btt_th,btt_dat,fisher)
       like(beamlike) = dlnlike_beam/2.d0
    endif

10  continue
    
  end SUBROUTINE act_likelihood_compute  

END MODULE act_likelihood_2008

